/*
 * fspy - Linux filesystem activity monitor
 *
 * Copyright (C) 2007  Richard Sammet (e-axe) <richard.sammet@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef _REGMATCH_H
#define _REGMATCH_H

int reg_comp(char *pattern);
int ireg_comp(char *pattern);
int reg_match(const char *string);
int ireg_match(const char *string);
void reg_dest(void);

#endif
