/*
 * fspy - Linux filesystem activity monitor
 *
 * Copyright (C) 2007  Richard Sammet (e-axe) <richard.sammet@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#include <ctype.h>

#include "fspy.h"
#include "isnumber.h"

int isnumber(char *nbrstr)
{

	int i = 0;

	while(nbrstr[i] != '\0') {
		if(!isdigit(nbrstr[i]))
			return FALSE;
		i++;
	}

	return TRUE;
}
