/*
 * fspy - Linux filesystem activity monitor
 *
 * Copyright (C) 2007  Richard Sammet (e-axe) <richard.sammet@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef _ADAPTIVE_H
#define _ADAPTIVE_H

#include <sys/stat.h>

#include "fspy.h"

void init_free_wds(unsigned int fdin);
int adaptive_check(int event_mask, const char *path, struct felement *lsptr,
                   struct stat *statdat, unsigned int wd);

#endif
