/*
 * fspy - Linux filesystem activity monitor
 *
 * Copyright (C) 2007  Richard Sammet (e-axe) <richard.sammet@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef _DIFF_H
#define _DIFF_H

#ifndef DELIM
#define DELIM ","
#endif

struct festat *init_diff(void);
struct festat *extend_diff_list(struct festat *felsptr);
void diffing(char *fpath, struct stat *statdat, struct diffprint *dprintptr,
             char *diffstring);

#endif
