/*
 * fspy - Linux filesystem activity monitor
 *
 * Copyright (C) 2007  Richard Sammet (e-axe) <richard.sammet@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef _OUTPUT_H
#define _OUTPUT_H

#include <sys/stat.h>
#include <sys/inotify.h>

#ifndef DELIM
#define DELIM ","
#endif

void print_data(char *mystrin, struct inotify_event *event,
                const char *event_fpath, struct stat *statdat, struct diffprint *dprint);

#endif
