/*
 * fanotify_events.h - fanotify-based event monitoring with process tracking
 *
 * Copyright (C) 2025 Bean Huo <beanhuo@micron.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifndef FANOTIFY_EVENTS_H
#define FANOTIFY_EVENTS_H

#include <sys/types.h>

/* Initialize fanotify and start monitoring the given path */
int fanotify_monitor(const char *path, char *output_fmt, char *filter,
                     char *ifilter, char *types, char *diff);

#endif /* FANOTIFY_EVENTS_H */
