/*
 * fspy - Linux filesystem activity monitor
 *
 * Copyright (C) 2007  Richard Sammet (e-axe) <richard.sammet@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#include "fspy.h"
#include "numlen.h"

int numlen(int nbr)
{

	int i = 0;

	while(nbr != 0) {
		nbr = nbr / 10;
		i++;
	}

	return i;
}
