/*
 * fspy - Linux filesystem activity monitor
 *
 * Copyright (C) 2007  Richard Sammet (e-axe) <richard.sammet@gmail.com>
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

#ifndef _ENUMDIRS_H
#define _ENUMDIRS_H

#define ELEMENT_SIZE              1024
#define ELEMENT_INIT_COUNT        1024
#define DIFF_ELEMENT_INIT_COUNT   2048

#define NO_DIR                    NULL

/* Forward declarations to avoid circular dependency */
struct felement;
struct stat;

int grab_max_element_count(void);
int pathlookup(char *lpath, unsigned int wd, struct felement *lsptr);
struct felement *recwrap(char *initial_path, int recursive_depth,
                         struct stat *statdat);

#endif
